#include "stdafx.h"

#include "bitmap.h"
#include "ff3se.h"

#include <png.h>

png_structp pngPtr;
png_infop infoPtr;
png_infop endInfo;

/*************************************/
//Loads a bitmap from filename (PNG file)
//Converts it to a 256 color paletted bitmap
//maxColors is the maximum number of colors the image can have
//maxWidth and maxHeight specify a maximum size for the image
//Returns 0 if successful, 1 if the image has over max colors, 2 if the image is too large,
//3 if an error occured, and 4 if it's not a valid PCX file
BYTE Bitmap::loadPNG(char *filename, int maxColors, int maxWidth, int maxHeight)
{
	int isPNG;
	BYTE header[8];
	FILE *fp;
	int bitDepth;
	int colorType;
	unsigned int w, h;
	int success;

	fp = fopen(filename, "rb");

	if (!fp)
		return 3;

	fread(header, 1, 5, fp);
	isPNG = !png_sig_cmp(header, 0, 5);		//Check to see if it's a valid PNG file
	if (!isPNG)
	{
		fclose(fp);
		return 4;
	}

	pngPtr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL,
		NULL, NULL);
	if (!pngPtr)
	{
		fclose(fp);
		return 3;
	}

	infoPtr = png_create_info_struct(pngPtr);
	if (!infoPtr)
	{
		png_destroy_read_struct(&pngPtr, NULL, NULL);
		fclose(fp);
		return 3;
	}

	endInfo = png_create_info_struct(pngPtr);
	if (!endInfo)
	{
		png_destroy_read_struct(&pngPtr, &infoPtr, NULL);
		fclose(fp);
		return 3;
	}

	png_init_io(pngPtr, fp);

	png_set_sig_bytes(pngPtr, 5);

	png_read_info(pngPtr, infoPtr);

	png_get_IHDR(pngPtr, infoPtr, &w, &h,
		&bitDepth, &colorType, NULL, NULL, NULL);

	width = (int)w;
	height = (int)h;

	if ((width > maxWidth) || (height > maxHeight))
	{
		png_destroy_read_struct(&pngPtr, &infoPtr, &endInfo);
		fclose(fp);
		return 2;
	}

	if (colorType == PNG_COLOR_TYPE_PALETTE)
	{
		success = loadPalPNG(maxColors, fp, bitDepth);
	}
	else
	{
		success = loadRGBPNG(maxColors, fp);
	}

	fclose(fp);

	png_destroy_read_struct(&pngPtr, &infoPtr, &endInfo);

	SAFE_DELETE(aBitmapBits);
	return success;
}

/*************************************/
//Loads data from a PNG image with a palette
//Calls convert256 to strip unused colors
//maxColors is the maximum number of colors the image can have
//Returns 0 if successful, 1 if the image has over max colors
BYTE Bitmap::loadPalPNG(int maxColors, FILE *fp, int bitDepth)
{
	BYTE *c;
	int x, y;
	BYTE *pixelPtr;
	int i;
	BYTE *trans;
	png_color_16 *transValues;
	int numTrans;
	png_color *pal;
	int numPal;
	int transColor;
	int found = -1;

	//Load palette info
	pal = (png_color *)png_malloc(pngPtr, sizeof(png_color) * 256);

	png_get_PLTE(pngPtr, infoPtr, &pal, &numPal);

	trans = (BYTE *)png_malloc(pngPtr, 256);
	transValues = (png_color_16 *)png_malloc(pngPtr, sizeof(png_color_16) * 256);

	for (i = 0; i < numPal; i++)
	{
		tempPal[i].rgbtBlue = pal[i].blue;
		tempPal[i].rgbtGreen = pal[i].green;
		tempPal[i].rgbtRed = pal[i].red;
	}

	if (bitDepth == 4)
	{
		//Copy 16 color palette
		for (int i = 0; i < 16; i++)
			palette[i] = tempPal[i];
	}

	//Check for transparency
	png_get_tRNS(pngPtr, infoPtr, &trans, &numTrans, &transValues);
	transColor = -1;
	for (i = 0; (i < numTrans) && (transColor == -1); i++)		//Determine which color is transparent
	{
		if (trans[i] == 0)
			transColor = i;
	}

	png_set_strip_alpha(pngPtr);		//Ignore the Alpha channel
	png_set_packing(pngPtr);			//Expand to 1 byte per pixel

	png_read_update_info(pngPtr, infoPtr);

	aBitmapBits = (BYTE *)malloc(width);

	pixels = (BYTE *)malloc(width * height);

	pixelPtr = pixels;

	if (bitDepth == 4)
		colorsUsed = 16;
	else
		colorsUsed = 0;

	for (y = 0; y < height; y++)			//Load pixel data
	{
		png_read_row(pngPtr, aBitmapBits, NULL);
		c = aBitmapBits;

		for (x = 0; x < width; x++)
		{
			if (convert256(tempPal[*c], x, y) > maxColors)
			{
				return	1;
			}
			c++;
		}
	}

	if (transColor != -1)			//If the PNG is transparent, set transparent color to palette entry 0
	{
		for (i = 0; (i < colorsUsed) && (found == -1); i++)
		{
			if ((palette[i].rgbtBlue == tempPal[transColor].rgbtBlue) &&
				(palette[i].rgbtGreen == tempPal[transColor].rgbtGreen) &&
				(palette[i].rgbtRed == tempPal[transColor].rgbtRed))
				found = i;
		}
		if (found != -1)
			swapPal(found, 0);
	}
	png_read_end(pngPtr, endInfo);
	return 0;
}

/*************************************/
//Loads data from a true color PNG
//Calls convert256 to strip unused colors and convert to 256 colors palette
//maxColors is the maximum number of colors the image can have
//Returns 0 if successful, 1 if the image has over max colors
BYTE Bitmap::loadRGBPNG(int maxColors, FILE *fp)
{
	BYTE *c;
	int x, y;
	BYTE *pixelPtr;
	int i;
	BYTE *trans;
	png_color_16 *transValues;
	int numTrans;
	RGBTRIPLE transColor;
	int found = -1;
	bool isTrans;
	RGBTRIPLE color;
	BYTE *bitmapData[1024];

	//Check for transparency
	isTrans = false;
	if (png_get_valid(pngPtr, infoPtr, PNG_INFO_tRNS) != 0)
	{
		png_get_tRNS(pngPtr, infoPtr, &trans, &numTrans, &transValues);

		if (numTrans > 0)
		{
			transColor.rgbtBlue = (BYTE)transValues[0].blue;
			transColor.rgbtGreen = (BYTE)transValues[0].green;
			transColor.rgbtRed = (BYTE)transValues[0].red;
			isTrans = true;
		}
	}

	png_set_strip_alpha(pngPtr);		//Ignore the Alpha channel

	png_read_update_info(pngPtr, infoPtr);

	for (i = 0; i < height; i++)
		bitmapData[i] = (BYTE *)malloc(width * 3);

	c = bitmapData[0];

	pixels = (BYTE *)malloc(width * height);

	pixelPtr = pixels;

	colorsUsed = 0;

	png_read_image(pngPtr, bitmapData);

	for (y = 0; y < height; y++)			//Load pixel data
	{
		c = (bitmapData[y]);

		for (x = 0; x < width; x++)
		{
			color.rgbtRed = *(c++);
			color.rgbtGreen = *(c++);
			color.rgbtBlue = *(c++);
			if (convert256(color, x, y) > maxColors)
			{
				return	1;
			}
		}
	}

	if (isTrans)			//If the PNG is transparent, set transparent color to palette entry 0
	{
		for (i = 0; (i < colorsUsed) && (found == -1); i++)
		{
			if ((palette[i].rgbtBlue == transColor.rgbtBlue) &&
				(palette[i].rgbtGreen == transColor.rgbtGreen) &&
				(palette[i].rgbtRed == transColor.rgbtRed))
				found = i;
		}
		if (found != -1)
			swapPal(found, 0);
	}

	for (i = 0; i < height; i++)
		SAFE_DELETE(bitmapData[i]);

	png_read_end(pngPtr, endInfo);
	return 0;
}

/*************************************/
//Saves the Bitmap to a file as a .PNG
BYTE Bitmap::savePNG(char *filename, int maxX, int maxY)
{
	FILE *fp;
	png_structp pngPtr;
	png_infop infoPtr;
	BYTE *RGBpixels[1024];
	int x, y;
	BYTE *ptr, *ptr2;
	int i;
	BYTE zero = 0;
	png_color pal[16];

	fp = fopen(filename, "wb");

	if (fp == NULL)
		return 1;


	pngPtr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

	if (pngPtr == NULL)
		return 1;

	infoPtr = png_create_info_struct(pngPtr);

	if (infoPtr == NULL)
	{
		png_destroy_write_struct(&pngPtr, NULL);
		return 1;
	}

	png_init_io(pngPtr, fp);

	//Set info header
	png_set_IHDR(pngPtr, infoPtr, maxX, maxY, 4, PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	//Set palette
	for (i = 0; i < colorsUsed; i++)
	{
		pal[i].red = palette[i].rgbtRed;
		pal[i].green = palette[i].rgbtGreen;
		pal[i].blue = palette[i].rgbtBlue;
	}
	png_set_PLTE(pngPtr, infoPtr, (png_color *)pal, colorsUsed);

	//Set transparency
	png_set_tRNS(pngPtr, infoPtr, &zero, 1, NULL);

	//Determine pixel data
	for (i = 0; i < maxY; i++)
		RGBpixels[i] = (BYTE *)malloc(maxX);

	for (y = 0; y < maxY; y++)
	{
		ptr = RGBpixels[y];
		ptr2 = pixels + (y * width);
		for (x = 0; x < maxX; x++)
		{
			*(ptr++) = *(ptr2++);
		}
	}

	//Write pixel data
	png_set_rows(pngPtr, infoPtr, RGBpixels);
	png_write_png(pngPtr, infoPtr, PNG_TRANSFORM_PACKING, NULL);

	png_write_end(pngPtr, infoPtr);

	fclose(fp);
	png_destroy_write_struct(&pngPtr, NULL);

	for (i = 0; i < maxY; i++)
		SAFE_DELETE(RGBpixels[i]);

	return 0;
}

